#include "stdafx.h"
#include "PluginMenu.h"

#include <msclr/marshal.h>
using namespace msclr::interop;

void __stdcall PWCMenuProc(TPWPInstance* instance, const char* item);
extern TPWPFunctions g_pwpfunc_t;

using namespace PWSPlugin;

PluginMenu::PluginMenu()
{
}

void PluginMenu::UpdateInstance(TPWPInstance* instance)
{
	this->instance = instance;
	for each (PluginMenuItem^ menu in this->items.Values)
	{
		menu->UpdateInstance(instance);
	}
}

PluginMenuItem^ PluginMenu::Add(String^ item, Style style, int groupId, PluginMenuItem::Proc^ proc)
{
	if (this->items.ContainsKey(item))
	{
		return nullptr;
	}

	marshal_context ctx;
	if (1 != g_pwpfunc_t.PWPMenuItem_Add(this->instance, ctx.marshal_as<const char*>(item), (int)style, groupId, PWCMenuProc))
	{
		return nullptr;
	}

	PluginMenuItem^ menu = gcnew PluginMenuItem(item, proc);
	this->items.Add(item, menu);
	return menu;
}

PluginMenuItem^ PluginMenu::Insert(String^ nextitem, String^ item, Style style, int groupId, PluginMenuItem::Proc^ proc)
{
	if (!this->items.ContainsKey(nextitem) || this->items.ContainsKey(item))
	{
		return nullptr;
	}

	marshal_context ctx;
	if (1 != g_pwpfunc_t.PWPMenuItem_Insert(this->instance, ctx.marshal_as<const char*>(nextitem), ctx.marshal_as<const char*>(item), (int)style, groupId, PWCMenuProc))
	{
		return nullptr;
	}

	PluginMenuItem^ menu = gcnew PluginMenuItem(item, proc);
	this->items.Add(item, menu);
	return menu;
}

PluginMenuItem^ PluginMenu::Rename(PluginMenuItem^ item, String^ newname)
{
	if (!this->items.ContainsKey(item->Name) || this->items.ContainsKey(newname))
	{
		return item;
	}

	marshal_context ctx;
	if (1 != g_pwpfunc_t.PWPMenuItem_Rename(this->instance, ctx.marshal_as<const char*>(item->Name), ctx.marshal_as<const char*>(newname)))
	{
		return item;
	}

	this->items.Remove(item->Name);
	item->Name = newname;
	this->items.Add(item->Name, item);
	return item;
}

bool PluginMenu::Delete(String^ item)
{
	if (!this->items.ContainsKey(item))
	{
		return false;
	}

	marshal_context ctx;
	if (1 != g_pwpfunc_t.PWPMenuItem_Delete(this->instance, ctx.marshal_as<const char*>(item)))
	{
		return false;
	}

	this->items.Remove(item);
	return true;
}

String^ PluginMenu::GetSetupMenuName()
{
	marshal_context ctx;
	const char* sel = g_pwpfunc_t.PWPMenuItem_GetSetupMenuName(this->instance);
	if (NULL == sel)
	{
		return nullptr;
	}

	return marshal_as<String^>(sel);
}

void PluginMenu::Call(String^ item)
{
	if (!this->items.ContainsKey(item))
	{
		System::Diagnostics::Debug::WriteLine("{0}: menu item {1} is not in items!", __FUNCTIONW__, item);
		return;
	}

	this->items[item]->Call();
}